import csv
import json
import os
import requests

from collections import UserDict

from sqlalchemy import create_engine, MetaData, Table
from sqlalchemy.orm import clear_mappers, mapper, sessionmaker


class Wrapper(UserDict):
    """Base class for creating dictionary wrappers.

    Defines a common __init__ and __setitem__ interface for child classes
    to implement."""
    def __init__(self, data=None, autosave=True, autorefresh=False,
                 *args, **kwargs):
        if data is None:
            data = {}
        self._autosave = autosave
        self._autorefresh = autorefresh
        self.data = self.load(data, *args, **kwargs)

    def load(self, data=None, update=None, *args, **kwargs):
        if data is None and hasattr(self, "data"):
            return self.data
        if update:
            data.update(update)
        data = {self.correct_key(key): self.correct_value(value)
                for key, value in data.items()}
        if update and self._autosave:
            self.save(data)
        return data

    def __setitem__(self, key, value):
        key = self.correct_key(key)
        value = self.correct_value(value)
        self.data[key] = value
        if self._autosave:
            self.pre_save_hook()
            self.save()

    def __getitem__(self, key):
        if self._autorefresh:
            self.refresh()
        return self.data[key]

    def save(self, data=None):
        pass

    def pre_save_hook(self):
        pass

    def refresh(self):
        self.data = self.load()

    def correct_key(self, key, *args, **kwargs):
        return key

    def correct_value(self, value, *args, **kwargs):
        return value


class json_dict(Wrapper):
    """Dict that reads from and writes to a JSON file."""
    def __init__(self, file_path=None, update=None, autosave=True,
                 autorefresh=False):
        if not file_path:
            raise AttributeError(
                "json_dict missing 1 required positional argument: file_path")
        self._file_path = file_path
        super().__init__(update, autosave, autorefresh)

    def load(self, update=None, *args, **kwargs):
        if not os.path.exists(self._file_path):
            return super().load({}, update)
        with open(self._file_path, "r") as f:
            json_data = json.load(f)
        return super().load(json_data, update)

    def save(self, data=None):
        with open(self._file_path, "w") as f:
            json.dump(data or self.data, f)

    def correct_key(self, key):
        try:
            key = int(key)
        except ValueError:
            pass
        return key


class DBRecord:

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)


class sql_row(Wrapper):
    """Dict that reads and writes to a SQL database row."""
    def __init__(self, database, table, lookup_field=None, lookup_id=None,
                 update=None, autosave=False, autorefresh=False):
        self._engine = create_engine(database)
        metadata = MetaData(self._engine)
        table = Table(table, metadata, autoload=True)
        self._mapper = mapper(DBRecord, table)
        Session = sessionmaker(bind=self._engine)
        self._session = Session()
        self._db_row = None
        self._updates = {}
        super().__init__(update, autosave, autorefresh,
                         lookup_field, lookup_id)

    def load(self, update=None, lookup_field=None, lookup_id=None):
        if self._db_row:
            self._session.refresh(self._db_row)
        elif not lookup_field and not lookup_id:
            self._db_row = self._create_row(update)
            update = None
        else:
            self._db_row = self._session.query(DBRecord).filter(
                getattr(DBRecord, lookup_field) == lookup_id).one()
        dict_data = self._get_dict_from_db_row()
        self._updates = update or {}
        return super().load(dict_data, update)

    def _get_dict_from_db_row(self):
        return {key: getattr(self._db_row, key)
                for key in dir(self._db_row) if not key.startswith("_")}

    def _create_row(self, data):
        db_record = DBRecord(**data)
        self._session.add(db_record)
        self._session.commit()
        self._session.refresh(db_record)
        return db_record

    def __setitem__(self, key, value):
        self._updates[key] = value
        super().__setitem__(key, value)

    def save(self, data=None):
        if data:
            self._updates = data
        for key, value in self._updates.items():
            setattr(self._db_row, key, value)
        self._updates = {}
        self._session.commit()

    def close(self):
        clear_mappers()
        self._session.close()


class csv_row(Wrapper):
    def __init__(self, file_path, row_id, fieldnames=None, update=None,
                 autosave=True, autorefresh=False):
        self._file_path = file_path
        self._file_data = []  # Set in load method.
        self._row_id = row_id
        self._fieldnames = fieldnames
        super().__init__(update, autosave, autorefresh)

    def load(self, update=None):
        with open(self._file_path, "r") as f:
            reader = csv.DictReader(f, fieldnames=self._fieldnames)
            if not self._fieldnames:
                self._fieldnames = reader.fieldnames
            self._file_data = [r for r in reader]
        if isinstance(self._row_id, int):
            row = self._file_data[self._row_id]
        elif isinstance(self._row_id, dict):
            self._row_id = self._get_row_by_field_value()
        row = self._file_data[self._row_id]
        return super().load(row, update)

    def _get_row_by_field_value(self):
        key, value = list(self._row_id.items())[0]
        for row_number, row_data in enumerate(self._file_data):
            if self.correct_value(row_data[key]) == value:
                return row_number

    def correct_key(self, key):
        return self.correct_value(key)

    def correct_value(self, value):
        if value.lower() == "true":
            return True
        elif value.lower() == "false":
            return False
        try:
            return int(value)
        except ValueError:
            return value

    def pre_save_hook(self):
        self._file_data[self._row_id] = self.data

    def save(self, data=None):
        with open(self._file_path, "w") as f:
            writer = csv.DictWriter(f, fieldnames=self._fieldnames)
            writer.writeheader()
            for row in self._file_data:
                writer.writerow(row)


class api_record(Wrapper):

    def __init__(self, endpoint, auth=None, update=None,
                 autosave=False, autorefresh=False):
        self._endpoint = endpoint
        self._auth = auth
        self._updates = {}
        super().__init__(update, autosave, autorefresh)

    def load(self, update=None):
        record = requests.get(self._endpoint, auth=self._auth).json()
        self._updates = update or {}
        return super().load(record, update)

    def save(self, data=None):
        requests.patch(self._endpoint, auth=self._auth,
                       json=json.dumps(self._updates))
        self._updates = {}

    def __setitem__(self, key, value):
        self._updates[key] = value
        super().__setitem__(key, value)
